from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import time, datetime

from dateutil import parser

TIME_FORMAT = '%H:%M'


class Model(ABC):

    @abstractmethod
    def to_payload(self) -> dict[str, any]:
        """Convert the model to a dictionary with strings as keys and any type as values."""

    @staticmethod
    @abstractmethod
    def from_payload(payload: dict[str, any]) -> 'Model':
        """Convert a dictionary with strings as keys values to a model."""


@dataclass
class UserModel(Model):
    data: dict | None = None

    def to_payload(self) -> dict[str, any]:
        payload = {}
        if self.data is not None:
            payload = self.data
        return payload

    @staticmethod
    def from_payload(payload: dict[str, any]) -> 'UserModel':
        return UserModel(
            data=payload
        )


@dataclass
class DeviceStateModel(Model):
    refill_time: time | None = None
    refill_interval_days: int | None = None
    max_pump_duration_minutes: int | None = None
    pump_state: bool | None = None

    def __post_init__(self):
        """Value validation after initialization"""
        if not isinstance(self.refill_time, time | None):
            raise ValueError('refill_time must be a time object or None')

        if not isinstance(self.refill_interval_days, int | None):
            raise ValueError('refill_interval_days must be an int or None')
        if self.refill_interval_days is not None and self.refill_interval_days <= 0:
            raise ValueError('refill_intervall must be positive')

        if not isinstance(self.max_pump_duration_minutes, int | None):
            raise ValueError('max_pump_duration_minutes must be an int or None')
        if self.max_pump_duration_minutes is not None and \
                not 0 < self.max_pump_duration_minutes < 24 * 60 - 1:
            raise ValueError('max_pump_duration_minutes must be between 0 and 1439 or None')

        if not isinstance(self.pump_state, bool | None):
            raise ValueError('pump_state must be a bool or None')

    def to_payload(self) -> dict[str, any]:
        payload = {}
        if self.refill_time is not None:
            payload['refillTime'] = self.refill_time.strftime('%H:%M')
        if self.refill_interval_days is not None:
            payload['refillInterval'] = f'{self.refill_interval_days}days'
        if self.max_pump_duration_minutes is not None:
            payload['maxPumpDuration'] = f'{self.max_pump_duration_minutes}min'
        if self.pump_state is not None:
            payload['pumpState'] = 'on' if self.pump_state else 'off'
        return payload

    @staticmethod
    def from_payload(payload: dict[str, any]) -> 'DeviceStateModel':
        return DeviceStateModel(
            refill_time=DeviceStateModel._parse_refill_time(payload),
            refill_interval_days=DeviceStateModel._parse_refill_interval(payload),
            max_pump_duration_minutes=DeviceStateModel._parse_max_pump_duration(payload),
            pump_state=DeviceStateModel._parse_pump_state(payload)
        )

    @staticmethod
    def _parse_max_pump_duration(payload) -> int | None:
        max_pump_duration_str = payload.get('maxPumpDuration')
        return max_pump_duration_str and int(max_pump_duration_str.replace('min', ''))

    @staticmethod
    def _parse_refill_time(payload) -> time | None:
        refill_time_str = payload.get('refillTime')
        return refill_time_str and datetime.strptime(refill_time_str, TIME_FORMAT).time()

    @staticmethod
    def _parse_refill_interval(payload) -> int | None:
        refill_interval_str = payload.get('refillInterval')
        return refill_interval_str and int(refill_interval_str.replace('days', ''))

    @staticmethod
    def _parse_pump_state(payload) -> bool | None:
        pump_state_str = payload.get('pumpState')
        match pump_state_str:
            case 'on':
                return True
            case 'off':
                return False
            case None:
                return None
            case _:
                raise ValueError(f'Unknown pump state {pump_state_str}')


@dataclass
class DeviceModel(Model):
    desired_state: DeviceStateModel | None = None
    reported_state: DeviceStateModel | None = None

    def to_payload(self) -> dict[str, any]:
        payload = {'state': {}}
        if self.desired_state:
            payload['state']['desired'] = self.desired_state.to_payload()
        if self.reported_state:
            payload['state']['reported'] = self.reported_state.to_payload()
        return payload

    @staticmethod
    def from_payload(payload: dict[str, any]) -> 'DeviceModel':
        desired_dict = payload.get('desired')
        desired = None if desired_dict is None else DeviceStateModel.from_payload(desired_dict)
        reported_dict = payload.get('reported')
        reported = None if desired_dict is None else DeviceStateModel.from_payload(reported_dict)

        return DeviceModel(desired, reported)


@dataclass
class DeviceDataModel(Model):
    data: dict[str, any]

    def __post_init__(self):
        """Value validation after initialization"""
        if not isinstance(self.data, dict):
            raise ValueError('data must be a dict')
        for k, v in self.data.items():
            if not isinstance(k, str):
                raise ValueError('data keys must be strings')
            if not isinstance(v, list):
                raise ValueError(
                    'data values must be lists of strings, numeric values, bools or date/time '
                    'objects')

    def to_payload(self) -> dict[str, any]:
        raise NotImplementedError('DeviceDataModel does not support to_payload')

    @staticmethod
    def from_payload(payload: dict[str, any]) -> 'DeviceDataModel':
        device_ids = DeviceDataModel._parse_device_ids(payload)
        timestamps = DeviceDataModel._parse_timestamps(payload)
        data = DeviceDataModel._parse_values(payload)
        data['deviceId'] = device_ids
        data['timestamp'] = timestamps
        return DeviceDataModel(data)

    @staticmethod
    def _parse_timestamps(payload: dict) -> list[datetime]:
        first_timeseries = list(payload['timeSeries'].values())[0]
        timestamps = [v['x'] for v in first_timeseries]
        return [parser.isoparse(t) for t in timestamps]

    @staticmethod
    def _parse_device_ids(payload: dict) -> list[str]:
        device_id = payload['details']['deviceId']
        first_timeseries = list(payload['timeSeries'].values())[0]
        return [device_id] * len(first_timeseries)

    @staticmethod
    def _parse_values(payload: dict) -> dict[str, list[any]]:
        def tryparse_float(value: str) -> float | None:
            try:
                return float(value)
            except TypeError:
                return None
            except ValueError:
                return None

        values = {}
        for name, data in payload['timeSeries'].items():
            values[name] = [tryparse_float(v['y']) for v in data]
        return values



from datetime import datetime, timedelta

from boum.api_client.v1.client import ApiClient
from boum.api_client.v1.models import DeviceStateModel, DeviceModel


class Device:
    # noinspection PyUnresolvedReferences
    """
        Abstration over the api client to easily interact with devices.

        Methods that get parts of the device state will return a tuple of the reported and desired
        state.

        Example
        -------
        >>> from datetime import time, datetime, timedelta
        >>> import pandas as pd
        >>> from boum.api_client.v1.client import ApiClient
        >>> from boum.resources.device import Device
        >>> from boum.api_client.v1.models import DeviceStateModel
        >>>
        >>> client = ApiClient(email, password, base_url=base_url)
        >>> # or ApiClient(refresh_token='token', base_url=base_url)
        >>>
        >>> with client:
        ...    # Get available device ids
        ...    device_ids = Device.get_device_ids(client)
        ...    # Create a device instance
        ...    device = Device(device_id, client)
        ...    # Remove device claim
        ...    # device.unclaim()
        ...    # Claim a device
        ...    # device.claim()
        ...    # Set desired device state
        ...    desired_device_State = DeviceStateModel(
        ...        pump_state=True,
        ...        refill_time=time(3, 32),
        ...        refill_interval_days=3,
        ...        max_pump_duration_minutes=5
        ...    )
        ...    device.set_desired_device_state(desired_device_State)
        ...    # Get reported and desired device state
        ...    reported, desired = device.get_device_states()
        ...    # Get device telemetry data
        ...    data = device.get_telemetry_data(start=datetime.now() - timedelta(days=1),
        ...        end=datetime.now())
        ...    # Convert telemetry data to pandas dataframe
        ...    df = pd.DataFrame(data)
        """

    def __init__(self, device_id: str, api_client: ApiClient):
        """
        Parameters
        ----------
            device_id
                The device id
            api_client
                The api client that handles the interaction with the api
        """
        self.device_id = device_id
        self._api_client = api_client

    @staticmethod
    def get_device_ids(api_client: ApiClient) -> list[str]:
        """Get all device ids

        Parameters
        ----------
            api_client
                The api client that handles the interaction with the api

        Returns
        -------
            list[str]
                The device ids
        """
        return api_client.root.devices.get()

    @staticmethod
    def get_claimed_device_ids(api_client: ApiClient) -> list[str]:
        """Get all claimed device ids

        Parameters
        ----------
            api_client
                The api client that handles the interaction with the api

        Returns
        -------
            list[str]
                The device ids
        """
        return api_client.root.devices.claimed.get()
    
    @staticmethod
    def get_device_details(api_client: ApiClient, 
                           only_claimed: bool = False, 
                           only_tested: bool = False, 
                           sku_contains: str = '',
                           created_after: datetime = None) -> list[dict]:
        """Get details of all claimed devices

        Parameters
        ----------
            api_client
                The api client that handles the interaction with the api

        Returns
        -------
            list[dict]
                The device details
        """
        all_devices = api_client.root.devices.get(include_details=True)
        claimed_devices = api_client.root.devices.claimed.get(include_details=True)
        claimed_devices = {device['id']: device for device in claimed_devices}
        device_details_list = []
        for device in all_devices:
            add_device = True
            if only_claimed and device['id'] not in list(claimed_devices.keys()):
                add_device = False
            elif sku_contains and sku_contains not in device['sku']:
                add_device = False
            elif only_tested and 'hasBeenTested' in device:
                if not device['hasBeenTested']:
                    add_device = False
            elif created_after and datetime.fromtimestamp(device['createdAt']['_seconds']) < created_after:
                add_device = False
            if add_device:
                if only_claimed:
                    device['claimedAt'] = datetime.fromtimestamp(claimed_devices[device['id']]['createdAt']['_seconds'])
                    device['ownerId'] = claimed_devices[device['id']]['ownerId']
                device_details_list.append(device)
        return device_details_list
    
    def set_desired_device_state(self, desired_device_state: DeviceStateModel):
        """
        Set the desired device state.

        Parameters
        ----------
            desired_device_state
                The desired device state
        """
        device_model = DeviceModel(desired_device_state)
        self._api_client.root.devices(self.device_id).patch(device_model)

    def get_device_states(self) -> (DeviceStateModel, DeviceStateModel):
        """
        Get the reported and desired device state.

        Returns
        -------
            a tuple with the reported and desired device states
        """
        device_model = self._api_client.root.devices(self.device_id).get()
        return device_model.reported_state, device_model.desired_state

    def get_telemetry_data(
            self, start: datetime = None, end: datetime = None,
            interval: timedelta = None) -> dict[str, list]:
        """
        Get telemetry data for a device

        Parameters
        ----------
            start
                The start date of the telemetry data
            end
                The end date of the telemetry data
            interval
                the interpolation interavl for the telemetry data

        Returns
        -------
            dict[str, list]
                The telemetry data in a format that can be easily converted to a pandas dataframe.
        """
        device_data_model = self._api_client.root.devices(self.device_id).data.get(
            start, end, interval)
        return device_data_model.data

    def claim(self, user_id: str = None):
        """
        Claim a device for the currently signed in use or a specified one.

        PArameters
        ----------
        user_id
            If this is specified, the device is claimed for the given user instead of the on that
            is signed in.
        """
        if user_id:
            self._api_client.root.devices(self.device_id).claim(user_id).put()
        else:
            self._api_client.root.devices(self.device_id).claim.put()

    def unclaim(self):
        """
        Remove any claim to the device.
        """
        self._api_client.root.devices(self.device_id).claim.delete()



import functools
import logging
from abc import ABC, abstractmethod
from typing import Callable

import requests
from requests import JSONDecodeError


class Endpoint(ABC):
    """Baseclass for all endpoint clients.

    With this class you can create a tree hierarchy of endpoints that models the topology of a
    REST API.

    The subclasses of this class should implement the methods that correspond to the http verbs that
    the endpoint supports, using the provided private methods.

    if there are nested endpoints, the subclasses should also have attributes that represent the
    nested paths. These should also be subclasses of EndpointClient.

    Only the arguments 'path_segment' and optionally 'disabled_for_collection' need to be set when
    defining child endpoint class attributes. The other arguments will be set automatically when the
    endpoint hierarchy tree is constructed. The root endpoint additionally needs to have the
    'session' and 'refresh_access_token' arguments set.
    """

    _access_token_expired_message = 'ExpiredAccessToken'  # nosec

    def __init__(
            self, path_segment: str, disabled_for_collection: bool = False,
            resource_id: str | None = None, session: requests.Session = None,
            refresh_access_token: Callable[[], None] = None, parent: 'Endpoint | None' = None):
        """
        Parameters
        ----------
            path_segment
                The path segment of the endpoint.
            disabled_for_collection
                If true, the endpoint will not be accessible if no resource id is provided.
            resource_id
                The id of the resource that this endpoint represents. If it is none, the enpoint
                represents a collection of resources.
            session
                The session to use for the requests.
            refresh_access_token
                A callable that, when executed, refreshes the access token.
            parent
                The parent endpoint. If it is none, the endpoint is the root of the tree.
        """
        self._path_segment = path_segment
        self._session = session
        self._refresh_access_token = refresh_access_token
        self._resource_id = resource_id
        self._disabled_for_collection = disabled_for_collection
        self._parent = parent

    @abstractmethod
    def __get__(self, parent, owner: type):
        """
        Validate attribute access and return a new instance of the attribute with a parent added.
        Every subclass must implement this method for propper type hinting with the propper
        return type as:
            def __get__(self, instance, owner: type) -> "...Endpoint":
                return super().__get__(instance, owner)
        """
        if isinstance(parent, Endpoint):
            if self._disabled_for_collection and parent.is_collection:
                raise AttributeError(
                    'This endpoint is only available for a single resource, not for a collection')
            return type(self)(
                path_segment=self._path_segment,
                disabled_for_collection=self._disabled_for_collection,
                session=parent.session,
                resource_id=self._resource_id,
                refresh_access_token=parent._refresh_access_token,
                parent=parent)

        return self

    def __call__(self, resource_id: str):
        """
        Returns a new endpoint of the same classe with an added resource id, representing a
        specific resource instead of a collection.
        """
        return type(self)(
            path_segment=self._path_segment,
            session=self._session,
            resource_id=resource_id,
            disabled_for_collection=self._disabled_for_collection,
            refresh_access_token=self._refresh_access_token,
            parent=self._parent)

    @property
    def session(self) -> requests.Session:
        """The session used for the requests."""
        return self._session

    @session.setter
    def session(self, value: requests.Session):
        self._session = value

    @property
    def url(self) -> str:
        """The full url of the endpoint."""
        path_elemets = [self._parent.url if self._parent else None,
                        self._path_segment,
                        self._resource_id]
        return '/'.join(s.strip('/') for s in path_elemets if s)

    @property
    def is_collection(self) -> bool:
        """True if the endpoint represents a collection of resources."""
        return self._resource_id is None

    @property
    def is_resource(self) -> bool:
        """True if the endpoint represents a single resource."""
        return not self.is_collection

    @property
    def _child_endpoints(self) -> list['Endpoint']:
        """Returns a list of all child endpoints."""
        return [e for e in vars(self) if isinstance(e, Endpoint)]

    # noinspection PyMethodParameters
    # pylint: disable=no-self-argument no-member protected-access
    def _request_handler(func: Callable[..., requests.Response]):
        """Decorator to apply the response handling method to the request methods"""

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):

            def execute_and_parse():

                if not self.session:
                    raise RuntimeError('Endpoint is not connected to the API')

                response = func(self, *args, **kwargs)
                try:
                    message = response.json().get('message')
                except JSONDecodeError:
                    message = 'No message'
                return response, message

            logging.info("Calling %s on %s...", func.__name__, self.url)
            response, message = execute_and_parse()

            if response.status_code == 401 \
                    and self._access_token_expired_message == message:
                logging.info('Access token expired. Refreshing...')
                self._refresh_access_token()
                logging.info('Access token refreshed. Retrying request...')
                response, message = execute_and_parse()

            if response.ok:
                logging.info('Request successful (%s): %s', response.status_code, message)
            else:
                logging.error('Request failed (%s): %s', response.status_code, message)
                response.raise_for_status()

            return response

        return wrapper

    # noinspection PyArgumentList
    @_request_handler
    def _get(self, payload: dict = None, query_parameters: dict = None):
        """Send a GET request to the endpoint."""
        return self._session.get(url=self.url, json=payload, params=query_parameters)

    # noinspection PyArgumentList
    @_request_handler
    def _post(self, payload: dict = None, query_parameters: dict = None):
        """Send a POST request to the endpoint."""
        return self._session.post(url=self.url, json=payload, params=query_parameters)

    # noinspection PyArgumentList
    @_request_handler
    def _put(self, payload: dict = None, query_parameters: dict = None):
        """Send a PUT request to the endpoint."""
        return self._session.put(url=self.url, json=payload, params=query_parameters)

    # noinspection PyArgumentList
    @_request_handler
    def _patch(self, payload: dict = None, query_parameters: dict = None):
        """Send a PATCH request to the endpoint."""
        return self._session.patch(url=self.url, json=payload, params=query_parameters)

    # noinspection PyArgumentList
    @_request_handler
    def _delete(self, payload: dict = None, query_parameters: dict = None):
        """Send a DELETE request to the endpoint."""
        return self._session.delete(url=self.url, json=payload, params=query_parameters)



import base64
from datetime import datetime, timedelta

import requests

from boum.api_client import constants
from boum.api_client.v1.endpoint import Endpoint
from boum.api_client.v1.models import DeviceModel, UserModel, DeviceDataModel


class ApiClient:
    # noinspection PyUnresolvedReferences
    """
        Client for the boum API v1.

        It is implemented as a context manager, so you can use it with
        the `with` statement. It will automatically connect and disconnect to the API. It will also
        automatically refresh the access token when it expires.

        A detailed documentation of the endpoint hierarchy can be found at the swagger page of
        the API (base_url/swagger).

        Attributes
        ----------
            root: EndpointClient
                The root endpoint client. It contains all the other nested endpoint clients.

        Example
        -------
            >>> from boum.api_client.v1.client import ApiClient
            >>> from boum.api_client.v1.models import DeviceModel
            >>>
            >>> client = ApiClient(email, password, base_url=base_url)
            >>> # or ApiClient(refresh_token='token', base_url=base_url)
            >>>
            >>> with client:
            ...     # Get call to the devices collection
            ...     device_ids = client.root.devices.get()
            ...     # Get call to a specific device
            ...     device_states = client.root.devices(device_id).get()
            ...     # Patch call to a specific device
            ...     client.root.devices(device_id).patch(DeviceModel())
            ...     # Get call to a devices data
            ...     data = client.root.devices(device_id).data.get()
        """

    def __init__(
            self, email: str = None, password: str = None, refresh_token: str = None, base_url:
            str = constants.API_URL_PROD, session: requests.Session = requests.Session()):
        """
        Parameters
        ----------
            email
                The email of the user. Required if `refresh_token` is not set.
            password
                The password of the user. Required if `refresh_token` is not set.
            refresh_token
                The refresh token of the user. Required if `email` and `password` are not set.
            base_url
                The URL of the API. Defaults to the production API.
        """
        self._email = None
        self.__access_token: str | None = None
        self.__refresh_token: bytes | None = None
        self.__password: bytes | None = None

        if email and password:
            self._email = email
            self._password = password
        elif refresh_token:
            self._refresh_token = refresh_token
        else:
            raise ValueError('Either email and password or refresh_token must be set')

        self._session = session
        self.root = RootEndpoint(base_url + '/v1', refresh_access_token=self._refresh_access_token)

    @property
    def _access_token(self) -> str | None:
        return self.__access_token

    @_access_token.setter
    def _access_token(self, value: str | None):
        self.__access_token = value
        self._session.headers.update({'Authorization': f'{self.__access_token}'})

    @property
    def _refresh_token(self) -> str | None:
        if self.__refresh_token:
            return base64.b64decode(self.__refresh_token).decode("utf-8")
        return None

    @_refresh_token.setter
    def _refresh_token(self, value: str | None):
        self.__refresh_token = base64.b64encode(value.encode('utf-8')) if value else None

    @property
    def _password(self) -> str | None:
        if self.__password:
            return base64.b64decode(self.__password).decode("utf-8")
        return None

    @_password.setter
    def _password(self, value: str | None):
        self.__password = base64.b64encode(value.encode('utf-8')) if value else None

    def __enter__(self) -> "ApiClient":
        """Connect to the API and sign in or refresh the access token."""
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Disconnect from the API."""
        self.disconnect()

    def disconnect(self):
        self._session.close()
        self.root.session = None

    def connect(self):
        self.root.session = self._session
        if self._access_token:
            pass
        elif self._refresh_token:
            self._refresh_access_token()
        else:
            self._signin()

    def _signin(self):
        self._access_token, self._refresh_token = self.root.auth.signin.post(
            self._email, self._password)

    def _refresh_access_token(self):
        if not self._refresh_token:
            raise AttributeError('Refresh token not set')

        self._access_token = self.root.auth.token.post(self._refresh_token)


class AuthTokenEndpoint(Endpoint):

    # pylint: disable=useless-parent-delegation
    def __get__(self, parent, owner: type) -> "AuthTokenEndpoint":
        return super().__get__(parent, owner)

    def post(self, refresh_token: str):
        if not isinstance(refresh_token, str):
            raise ValueError('refresh_token must be a string')

        payload = {'refreshToken': refresh_token}
        response = self._post(payload)
        data = response.json()['data']
        return data['accessToken']


class AuthSigninEndpoint(Endpoint):

    # pylint: disable=useless-parent-delegation
    def __get__(self, parent, owner: type) -> "AuthSigninEndpoint":
        return super().__get__(parent, owner)

    def post(self, email: str, password: str):
        if not isinstance(email, str):
            raise ValueError('email must be a string')
        if not isinstance(password, str):
            raise ValueError('password must be a string')

        payload = {'email': email, 'password': password}
        response = self._post(payload)
        data = response.json()['data']
        return data['accessToken'], data['refreshToken']


class AuthEndpoint(Endpoint):
    signin = AuthSigninEndpoint('signin')
    token = AuthTokenEndpoint('token')

    # pylint: disable=useless-parent-delegation
    def __get__(self, parent, owner: type) -> "AuthEndpoint":
        return super().__get__(parent, owner)


class DevicesDataEndpoint(Endpoint):
    DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%SZ'

    # pylint: disable=useless-parent-delegation
    def __get__(self, parent, owner: type) -> "DevicesDataEndpoint":
        return super().__get__(parent, owner)

    def get(self, start: datetime = None, end: datetime = None, interval: timedelta = None):
        if self._parent.is_collection:
            raise AttributeError('Cannot get data for a collection of devices')
        if start is not None and not isinstance(start, datetime):
            raise ValueError('start must be a datetime')
        if end is not None and not isinstance(end, datetime):
            raise ValueError('end must be a datetime')
        if interval is not None and not isinstance(interval, timedelta):
            raise ValueError('interval must be a timedelta')

        query_parameters = {}
        if start:
            query_parameters['timeStart'] = start.strftime(self.DATETIME_FORMAT)
        if end:
            query_parameters['timeEnd'] = end.strftime(self.DATETIME_FORMAT)
        if interval:
            interval_seconds = int(interval.total_seconds())
            query_parameters['interval'] = f'{interval_seconds}s'

        response = self._get(query_parameters=query_parameters)
        return DeviceDataModel.from_payload(response.json()['data'])


class DevicesClaimEndpoint(Endpoint):

    # pylint: disable=useless-parent-delegation
    def __get__(self, parent, owner: type) -> "DevicesClaimedEndpoint":
        return super().__get__(parent, owner)

    def put(self):
        self._put()

    def delete(self):
        if self.is_resource:
            raise AttributeError('Cannot unclaim from a specific user')
        self._delete()

class DevicesClaimedEndpoint(Endpoint):

    # pylint: disable=useless-parent-delegation
    def __get__(self, parent, owner: type) -> "DevicesClaimEndpoint":
        return super().__get__(parent, owner)

    def get(self, include_details: bool = False) -> list[str | dict] | DeviceModel:
        response = self._get()
        data = response.json()['data']
        if self.is_collection:
            if include_details:
                return data
            return [d['id'] for d in data]

        device_model = DeviceModel.from_payload(data)
        return device_model


class DevicesEndpoint(Endpoint):
    data = DevicesDataEndpoint('data')
    claim = DevicesClaimEndpoint('claim')
    claimed = DevicesClaimedEndpoint('claimed')

    # pylint: disable=useless-parent-delegation
    def __get__(self, parent, owner: type) -> "DevicesEndpoint":
        return super().__get__(parent, owner)

    def post(self) -> str:
        if self.is_resource:
            raise ValueError('Cannot post to a specific device')
        response = self._post()
        data = response.json()['data']
        return data['deviceId']

    def get(self, include_details: bool = False) -> list[str | dict] | DeviceModel:
        response = self._get()
        data = response.json()['data']
        if self.is_collection:
            if include_details:
                return data
            return [d['id'] for d in data]

        device_model = DeviceModel.from_payload(data)
        return device_model

    def patch(self, device_model: DeviceModel):
        if self.is_collection:
            raise ValueError('Cannot patch a collection of devices')
        if not isinstance(device_model, DeviceModel):
            raise ValueError('device_model must be a DeviceModel')

        payload = device_model.to_payload()
        self._patch(payload)

    def delete(self):
        if self.is_collection:
            raise ValueError('Cannot delete a collection of devices')
        raise NotImplementedError()


class UsersEndpoint(Endpoint):

    # pylint: disable=useless-parent-delegation
    def __get__(self, parent, owner: type) -> "UsersEndpoint":
        return super().__get__(parent, owner)

    def get(self) -> UserModel:
        response = self._get()
        payload = response.json()['data']
        return UserModel.from_payload(payload)


class RootEndpoint(Endpoint):
    devices = DevicesEndpoint('devices')
    auth = AuthEndpoint('auth')
    users = UsersEndpoint('users')

    # pylint: disable=useless-parent-delegation
    def __get__(self, parent, owner: type) -> "RootEndpoint":
        return super().__get__(parent, owner)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['boum', 'boum.api_client', 'boum.api_client.v1', 'boum.resources']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.28.1,<3.0.0']

setup_kwargs = {
    'name': 'boum',
    'version': '1.1.1',
    'description': 'SDK to interact with the boum API',
    'long_description': "# Boum Python SDK\nStatus: ![GitHub Actions](https://github.com/boum-garden/sdk/actions/workflows/deploy_on_semver_tag.yml/badge.svg)\n\n## Overview\nThis project provides an api client to interact with the [Boum API](https://api.boum.us/swagger) and abstractions to \nsimplify interactions with the underlying resources.\n\n\n## Installation\nThe package is available on PyPI, and can be installed with pip or similar tools:\n\n```bash\n    python -m pip install boum\n    poetry add boum\n    pipenv install boum\n    ...\n```\n\n## Usage\n\n### API Client\nThe API client models the topology of the API and provides a class hierarchy that is organized in the same way as the \nendpoint paths. Email and password or a token are required to use it.\n\nThe client can either connect to the API with `client.connect()` and `client.disconnect()`methods or it can be used \nas a context manager as `with client: ...` In the former case the disconnect method should be called explicitly in the\nend.\n\n```python\n>>> from boum.api_client.v1.client import ApiClient\n>>> from boum.api_client.v1.models import DeviceModel\n>>> \n>>> client = ApiClient(email, password, base_url=base_url)\n>>> # or ApiClient(refresh_token='token', base_url=base_url)\n>>>\n>>> with client:\n...     # Get call to the devices collection\n...     device_ids = client.root.devices.get()\n...     # Get call to a specific device \n...     device_states = client.root.devices(device_id).get()\n...     # Patch call to a specific device\n...     client.root.devices(device_id).patch(DeviceModel())\n...     # Get call to a devices data\n...     data = client.root.devices(device_id).data.get()\n\n```\n\n\n### Resource Abstractions\nThe resource abstractions provide a more intuitive interface to interact with the underlying resources.\n\n```python\n>>> from datetime import time, datetime, timedelta\n>>> import pandas as pd\n>>> from boum.api_client.v1.client import ApiClient\n>>> from boum.resources.device import Device\n>>> from boum.api_client.v1.models import DeviceStateModel\n>>>\n>>> client = ApiClient(email, password, base_url=base_url)\n>>> # or ApiClient(refresh_token='token', base_url=base_url)\n>>>\n>>> with client:\n...    # Get available device ids\n...    device_ids = Device.get_device_ids(client)\n...    # Create a device instance\n...    device = Device(device_id, client)\n...    # Remove device claim\n...    # device.unclaim()\n...    # Claim a device\n...    # device.claim()\n...    # Set desired device state\n...    desired_device_State = DeviceStateModel(\n...        pump_state=True,\n...        refill_time=time(3, 32),\n...        refill_interval_days=3,\n...        max_pump_duration_minutes=5\n...    )\n...    device.set_desired_device_state(desired_device_State)\n...    # Get reported and desired device state\n...    reported, desired = device.get_device_states()\n...    # Get device telemetry data\n...    data = device.get_telemetry_data(start=datetime.now() - timedelta(days=1),\n...        end=datetime.now())\n...    # Convert telemetry data to pandas dataframe\n...    df = pd.DataFrame(data)\n\n```\n\n### Jupyter Notebook Demo\nA Jupyter notebook demo is available [here](https://github.com/boum-garden/sdk/blob/638d62836f5b8b2f169d186170c679d6a813867a/doc/boum-sdk-demo.ipynb).\n\n### Loging\nThe SDK uses the standard python logging module.\n\n## Development\n\n### Dependecy management\n[Poetry](https://python-poetry.org/) is used for depency management and publishing.\n\n### Versioning\nVersioning is done using [bumpver](https://pypi.org/project/bumpver/) \nwith [semantic versioning](https://semver.org/)\nWhen the version is updated, a new tag is created and pushed to the remote repository.\n\n\n### Testing\n\n#### Unit tests\nThese are completely self-contained and have no external dependencies.\n\nIMPORTANT: There are fixtures that model the API responses and expected calls. These mnust be kept up to date with\nthe API, otherwise the unit tests will not test the correct behavior.\n\n#### End-to-end tests\nThese test the entire flow from user facing python classes to the underlying api calls. They require an instance \n(fake or real) of the API to run against. They also require a registered user with a clamied device.\nAPI Base URL, email and password are required as environmental variables.\n\n```bash\n    BOUM_SDK_TEST_BASE_URL\n    BOUM_SDK_TEST_EMAIL\n    BOUM_SDK_TEST_PASSWORD\n    BOUM_SDK_TEST_DEVICE_ID\n    BOUM_SDK_TEST_USER_ID\n```\n\n#### Doctests\n\nPart of the end-to-end tests are document tests executed using \n[doctest](https://docs.python.org/3/library/doctest.html). \nThese tests ensure that all the examples in this README and in the docstrings are up-to-date and working.\n\n\n## CI/CD\n\n### Integration\n\nThe GitHub action workflow `.github/workflows/checks_on_push_to_branch.yml` is triggered on every push on a branch\nexcept main. It provides the following checks:\n\n- Unit and end-to-end/contract tests with pytest\n- Linting with pylint\n- Security checks with bandit\n\nThe repository settings require that all checks pass before a pull request can be merged.\n\n### Deployment\n\nThe GitHub action workflow `.github/workflows/deploy_on_push_to_main.yml` is triggered on every push to main. It runs a\ntest deployment to the test PyPI repository.\n\nThe GitHub action workflow `.github/workflows/deploy_on_release.yml` is triggered on a tag with the format `v*.*.*` and\nruns a deployment to the production PyPI. It is recommended to set these tags using the GitHub release feature. \n\n\n## Technical debt\n\n- Doctest requires a specific format to make examples in documentation executable. This makes the examples in this\n  README harder to copy and paste. There exist other packages and add-ons that could be used to make the examples easier \n  to manage.\n\n\n\n\n\n\n",
    'author': 'Marc Pfander',
    'author_email': 'marc.pfander@boum.garden',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/boum-garden/sdk',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
